from __future__ import division

import functools
import time
import warnings
from typing import List, Optional, Tuple, Union

import numpy as np
from scipy import linalg, sparse
from scipy.optimize import LinearConstraint, fmin_l_bfgs_b, minimize
from sklearn.exceptions import ConvergenceWarning
from sklearn.utils import check_array
from sklearn.utils.validation import check_random_state

from ._cd_fast import (
    _norm_min_subgrad,
    enet_coordinate_descent_gram,
    identify_active_rows,
)
from ._distribution import ExponentialDispersionModel
from ._link import Link
from ._util import _safe_lin_pred, _safe_sandwich_dot


def timeit(runtime_attr: str):
    """
    Decorate a function to compute its run time and update the \
    :class:`IRLSState` instance attribute called ``runtime_attr`` with the run \
    time of the function.

    The first argument of ``fct`` should be an :class:`IRLSState` instance.
    """

    def fct_wrap(fct):
        @functools.wraps(fct)
        def inner_fct(*args, **kwargs):
            start = time.perf_counter()
            out = fct(*args, **kwargs)
            setattr(args[0], runtime_attr, time.perf_counter() - start)
            return out

        return inner_fct

    return fct_wrap


@timeit("inner_solver_runtime")
def _least_squares_solver(state, data, hessian):
    if data.has_lower_bounds or data.has_upper_bounds:
        raise ValueError("Bounds are not supported with the least squares solver.")

    # TODO: In cases where we have lots of columns, we might want to avoid the
    # sandwich product and use something like iterative lsqr or lsmr.
    d = linalg.solve(hessian, state.score, assume_a="pos")
    return d, 1


@timeit("inner_solver_runtime")
def _cd_solver(state, data, active_hessian):
    new_coef, gap, _, _, n_cycles = enet_coordinate_descent_gram(
        state.active_set,
        state.coef.copy(),
        data.P1,
        active_hessian,
        -state.score,
        data.max_inner_iter,
        state.inner_tol,
        data.random_state,
        data.fit_intercept,
        data.selection == "random",
        data.has_lower_bounds,
        data._lower_bounds,
        data.has_upper_bounds,
        data._upper_bounds,
    )
    return new_coef - state.coef, n_cycles


@timeit("build_hessian_runtime")
def update_hessian(state, data, active_set):
    """
    Update the approximate Hessian.

    The goal is to compute ``H = X^T @ diag(hessian_rows) @ X``. We will refer
    to ``H`` as the Hessian, even though technically we are computing a
    Gauss-Newton approximation to the Hessian.

    Instead of computing ``H`` directly, we will compute updates to ``H``:
    ``dH``. So, given ``H0`` from a previous iteration,
    ``H0 = X^T @ diag(hessian_rows_0) @ X``, we want to compute ``H1`` from this
    iteration: ``H1 = X^T @ diag(hessian_rows_1) @ X``.

    However, we will instead compute ``H1 = H0 + dH``, where
    ``dH = X^T @ diag(hessian_rows_1 - hessian_rows_0) @ X``.

    We will also refer to
    ``hessian_rows_diff = hessian_rows_1 - hessian_rows_0``.

    The advantage of reframing the computation of ``H`` as an update is that the
    values in ``hessian_rows_diff`` will vary depending on how large the
    influence of that last coefficient update was on that row. As a result, in
    lots of problems, many of the entries in ``hessian_rows_diff`` will be very
    very small.

    So, the goal with ``identify_active_rows`` is to filter to a subset of
    ``hessian_rows_diff`` that we will use to compute the sandwich product for
    ``dH``. If ``threshold/data.hessian_approx == 0.0``, then we will always use
    every row. However, for ``data.hessian_approx != 0``, we include rows for
    which
    ``include = (np.abs(hessian_rows_diff[i]) >= T * np.max(np.abs(hessian_rows_diff)))``.

    Essentially, this criterion ignores data matrix rows that have not seen the
    second derivatives of their predictions change very much in the last
    iteration.

    Critically, we set
    ``hessian_rows_old[include] += hessian_rows_diff[include]``.
    That way, ``hessian_rows_diff`` is no longer the change since the last
    iteration, but, instead, the change since the last iteration that a row was
    active. This ensures that we don't miss the situation where a row changes a
    small amount over several iterations which accumulates into a large change.
    """
    # The simplest, but expensive option is that if the active set increases in
    # size, we can just recompute the full hessian for the new active set
    # instead of an approximate update. This is what is currently implemented
    # here.
    #
    # Another interesting option would be to keep track of two old sets of
    # hessian_rows: one for the full dataset and one for the current reduced
    # dataset. Restart from the full dataset version if we increase the size of
    # the active_set. Though expensive, that would be a rare operation.
    #
    # Third thing: we could have a flag that swaps between computing H and
    # delta H. If we just computed H directly, then this wouldn't be an issue.
    # A slight modification of this: we could still use a baseline H from the
    # first iteration with the entire column set.
    first_iteration = not state.hessian_initialized
    reset_iteration = not _is_subset(state.old_active_set, active_set)
    if first_iteration or reset_iteration:

        # In the first iteration or in a reset iteration, we need to:
        # 1) use hessian_rows, not the difference
        # 2) use all the rows
        # 3) Include the P2 components
        # 4) just like an update, we only update the active_set
        hessian_init = build_hessian_delta(
            data.X,
            state.hessian_rows,
            data.fit_intercept,
            data.P2,
            np.arange(data.X.shape[0], dtype=np.int32),
            active_set,
        )
        state.hessian[np.ix_(active_set, active_set)] = hessian_init
        state.hessian_initialized = True
        n_active_rows = data.X.shape[0]
    else:

        # In an update iteration, we want to:
        # 1) use the difference in hessian_rows from the last iteration
        # 2) filter for active_rows in case data.hessian_approx != 0
        # 3) Ignore the P2 components because those don't change and have
        #    already been added
        # 4) only update the active set subset of the hessian.
        hessian_rows_diff, active_rows = identify_active_rows(
            state.gradient_rows,
            state.hessian_rows,
            state.old_hessian_rows,
            data.hessian_approx,
        )
        hessian_delta = build_hessian_delta(
            data.X,
            hessian_rows_diff,
            data.fit_intercept,
            P2=None,
            active_rows=active_rows,
            active_cols=active_set,
        )
        state.hessian[np.ix_(active_set, active_set)] += hessian_delta
        n_active_rows = active_rows.shape[0]

    return (
        state.hessian[np.ix_(active_set, active_set)],
        n_active_rows,
    )


def _is_subset(x, y):
    # NOTE: This functions assumes entries in x and y are unique
    intersection = np.intersect1d(x, y)
    return intersection.size == y.size


def build_hessian_delta(
    X, hessian_rows, intercept, P2, active_rows, active_cols
) -> np.ndarray:
    """
    Get the Hessian "sandwich" matrix for active rows and columns.

    Parameters
    ----------
    X
    hessian_rows
    intercept
    P2
    active_rows
    active_cols
    """
    idx = 1 if intercept else 0
    active_cols_non_intercept = active_cols[idx:] - idx
    delta = _safe_sandwich_dot(
        X, hessian_rows, active_rows, active_cols_non_intercept, intercept
    )
    if P2 is None:
        return delta

    if sparse.issparse(P2) and P2.nnz == 0:
        return delta

    if P2.ndim == 1:
        idiag = np.arange(start=idx, stop=delta.shape[0])
        delta[(idiag, idiag)] += P2[active_cols_non_intercept]
    else:
        if sparse.issparse(P2):
            is_diagonal = P2.nnz == P2.shape[0] and (P2.data == P2.diagonal()).all()
            if is_diagonal:
                idiag = np.arange(start=idx, stop=delta.shape[0])
                delta[(idiag, idiag)] += P2.data[active_cols_non_intercept]
            else:
                delta[idx:, idx:] += P2.toarray()[
                    np.ix_(active_cols_non_intercept, active_cols_non_intercept)
                ]
        else:
            delta[idx:, idx:] += P2[
                np.ix_(active_cols_non_intercept, active_cols_non_intercept)
            ]
    return delta


def _irls_solver(inner_solver, coef, data) -> Tuple[np.ndarray, int, int, List[List]]:
    """
    Solve GLM with L1 and L2 penalty by IRLS.

    The objective being minimized in the coefficients w=coef is::

        F = f + g, f(w) = 1/2 deviance, g = 1/2 w*P2*w + ||P1*w||_1

    An Improved GLMNET for L1-regularized Logistic Regression:

    1. Find optimal descent direction d by minimizing
       min_d F(w+d) = min_d F(w+d) - F(w)
    2. Quadratic approximation of F(w+d)-F(w) = q(d):
       using f(w+d) = f(w) + f'(w)*d + 1/2 d*H(w)*d + O(d^3) gives:
       q(d) = (f'(w) + w*P2)*d + 1/2 d*(H(w)+P2)*d
       + ||P1*(w+d)||_1 - ||P1*w||_1
       Then minimize q(d): min_d q(d)
    3a. Coordinate descent by updating coordinate j (d -> d+z*e_j):
       min_z q(d+z*e_j)
       = min_z q(d+z*e_j) - q(d)
       = min_z A_j z + 1/2 B_jj z^2
               + ||P1_j (w_j+d_j+z)||_1 - ||P1_j (w_j+d_j)||_1
       A = f'(w) + d*H(w) + (w+d)*P2
       B = H + P2
    3b. Least squares solve of the quadratic approximation.

    Repeat steps 1-3 until convergence.
    Note: Use hessian matrix instead of Hessian for H.
    Note: f' = -score, H = hessian matrix

    Parameters
    ----------
    inner_solver
        A least squares solver that can handle the appropriate penalties. With
        an L1 penalty, this will _cd_solver. With only an L2 penalty,
        _least_squares_solver will be more efficient.
    coef : ndarray, shape (c,)
        If fit_intercept=False, shape c=X.shape[1].
        If fit_intercept=True, then c=X.shape[1] + 1.
    data : IRLSData
        Data object containing all the data and solver parameters.

    References
    ----------
    Guo-Xun Yuan, Chia-Hua Ho, Chih-Jen Lin
    An Improved GLMNET for L1-regularized Logistic Regression,
    Journal of Machine Learning Research 13 (2012) 1999-2030
    https://www.csie.ntu.edu.tw/~cjlin/papers/l1_glmnet/long-glmnet.pdf
    """
    state = IRLSState(coef, data)

    state.eta, state.mu, state.obj_val, coef_P2 = _update_predictions(
        state, data, state.coef
    )
    state.gradient_rows, state.score, state.hessian_rows = update_quadratic(
        state, data, coef_P2
    )
    (
        state.converged,
        state.norm_min_subgrad,
        state.max_min_subgrad,
        state.inner_tol,
    ) = check_convergence(state, data)

    state._record_iteration()

    with ProgressBar(state.norm_min_subgrad, data.gradient_tol, data.verbose) as pb:
        while state.n_iter < data.max_iter and not state.converged:
            pb._update(state.n_iter, state.iteration_runtime, state.norm_min_subgrad)

            state.old_active_set = state.active_set
            state.active_set = identify_active_set(state, data)

            # 0) Build the hessian
            hessian, state.n_active_rows = update_hessian(state, data, state.active_set)

            # 1) Solve the L1 and L2 penalized least squares problem
            d, n_cycles_this_iter = inner_solver(state, data, hessian)
            state.n_cycles += n_cycles_this_iter

            # 2) Line search
            state.old_hessian_rows[:] = state.hessian_rows
            (
                state.coef,
                state.step,
                state.eta,
                state.mu,
                state.obj_val,
                coef_P2,
            ) = line_search(state, data, d)
            state.n_updated = np.sum(np.abs(d) > 0)

            # 3) Update the quadratic approximation
            state.gradient_rows, state.score, state.hessian_rows = update_quadratic(
                state, data, coef_P2
            )

            # 4) Check if we've converged
            (
                state.converged,
                state.norm_min_subgrad,
                state.max_min_subgrad,
                state.inner_tol,
            ) = check_convergence(state, data)

            state._record_iteration()

    if not state.converged:
        warnings.warn(
            "IRLS failed to converge. Increase"
            " the maximum number of iterations max_iter"
            " (currently {})".format(data.max_iter),
            ConvergenceWarning,
        )
    return state.coef, state.n_iter, state.n_cycles, state.diagnostics


class ProgressBar:
    """
    Display the current progress of the solver using tqdm.

    Current format is that:
        * The "Iteration #" is the IRLS iteration #.
        * The bar itself measures the log base 10 progress towards the
            gradient_tol. So, if we start at gradient_norm = 10 and are going to
            gradient_norm = 1e-4, the progress bar will be out of 5. If the
            current gradient_norm is 0.1, then we will be at 2/5.
        * On the right, show the time for the most recent iteration and the
            current gradient norm.
    """

    def __init__(self, start_norm, tol, verbose):
        self.start_norm = start_norm
        self.tol = tol
        self.verbose = verbose

    def __enter__(self):
        """Run the tqdm progress bar."""
        if not self.verbose:
            return self
        bar_start_loggrad = np.log10(self.start_norm)
        bar_end_loggrad = np.log10(self.tol)
        self.n_bar_steps = np.ceil(bar_start_loggrad - bar_end_loggrad)
        # Wait to import so that if verbose=False, we don't need tqdm installed
        from tqdm import tqdm

        self.t = tqdm(
            total=self.n_bar_steps,
            bar_format="Iteration {postfix[0]}: {l_bar}{bar}| {n_fmt}/{total_fmt} [{pos"
            "tfix[1]}s/it, gradient norm={postfix[2]}]",
            postfix=[0, "", self.start_norm],
        )
        return self

    def __exit__(self, *exc):
        """Stop tracking progress."""
        if self.verbose:
            self.t.close()

    def _update(self, n_iter, iteration_runtime, cur_grad_norm):
        if not self.verbose:
            return
        self.t.postfix[0] = n_iter
        self.t.postfix[1] = f"{iteration_runtime:.2f}"
        self.t.postfix[2] = cur_grad_norm
        # clip to 0 in case we take a step in the wrong direction at the start
        # without this, tqdm will print an annoying warning.
        step = max(self.n_bar_steps - (np.log10(cur_grad_norm) - np.log10(self.tol)), 0)
        # round to two digits for beauty
        self.t.n = np.round(step, 2)
        self.t.update(0)


class IRLSData:
    """Store parameters for the IRLS optimizer."""

    def __init__(
        self,
        X,
        y: np.ndarray,
        sample_weight: np.ndarray,
        P1: Union[np.ndarray, sparse.spmatrix],
        P2: Union[np.ndarray, sparse.spmatrix],
        fit_intercept: bool,
        family: ExponentialDispersionModel,
        link: Link,
        max_iter: int = 100,
        max_inner_iter: int = 100000,
        gradient_tol: Optional[float] = 1e-4,
        step_size_tol: Optional[float] = 1e-4,
        hessian_approx: float = 0.0,
        fixed_inner_tol: Optional[Tuple] = None,
        selection="cyclic",
        random_state=None,
        offset: Optional[np.ndarray] = None,
        lower_bounds: Optional[np.ndarray] = None,
        upper_bounds: Optional[np.ndarray] = None,
        verbose: bool = False,
    ):
        self.X = X
        self.y = y
        self.sample_weight = sample_weight
        self.P1 = P1

        # Note: we already set P2 = l2*P2, P1 = l1*P1
        # Note: we already symmetrized P2 = 1/2 (P2 + P2')
        self.P2 = P2

        self.fit_intercept = fit_intercept
        self.family = family
        self.link = link
        self.max_iter = max_iter
        self.max_inner_iter = max_inner_iter
        self.gradient_tol = gradient_tol
        self.step_size_tol = step_size_tol
        self.hessian_approx = hessian_approx
        self.fixed_inner_tol = fixed_inner_tol
        self.selection = selection
        self.random_state = random_state
        self.offset = offset
        self.has_lower_bounds, self._lower_bounds = _setup_bounds(
            lower_bounds, self.X.dtype
        )
        self.has_upper_bounds, self._upper_bounds = _setup_bounds(
            upper_bounds, self.X.dtype
        )

        self.intercept_offset = 1 if self.fit_intercept else 0
        self.verbose = verbose

        self._check_data()

    def _check_data(self):
        if self.P2.ndim == 2:
            self.P2 = check_array(self.P2, "csc", dtype=[np.float64, np.float32])

        if sparse.issparse(self.X):
            if not sparse.isspmatrix_csc(self.P2):
                raise ValueError(
                    "If X is sparse, P2 must also be sparse CSC format. It is not."
                )

        self.random_state = check_random_state(self.random_state)


def _setup_bounds(
    bounds: Optional[np.ndarray], dtype
) -> Tuple[bool, Optional[np.ndarray]]:
    _out_bounds = bounds
    if _out_bounds is None:
        _out_bounds = np.array([], dtype=dtype)
    return bounds is not None, _out_bounds


class IRLSState:
    """Track many parameters, such as score and Hessian, used in the IRLS solver."""

    def __init__(self, coef, data):
        self.data = data

        # some precalculations
        self.iteration_start = time.time()
        self.iteration_runtime = 0.0

        # number of outer iterations
        self.n_iter = -1

        # number of inner iterations (for CD, this is the number of cycles over
        # all the features)
        self.n_cycles = 0

        self.converged = False

        self.diagnostics = []

        self.coef = coef

        # We need to have an initial step value to make sure that the step size
        # convergence criteria fails on the first pass
        initial_step = data.step_size_tol
        if initial_step is None:
            initial_step = 0.0
        self.step = np.full_like(self.coef, initial_step)

        self.obj_val = None
        self.eta = np.zeros(data.X.shape[0], dtype=data.X.dtype)
        self.mu = None
        self.score = np.empty_like(self.coef)
        self.old_hessian_rows = np.zeros(data.X.shape[0], dtype=data.X.dtype)
        self.gradient_rows = None
        self.hessian_rows = None
        self.hessian = np.zeros(
            (self.coef.shape[0], self.coef.shape[0]), dtype=data.X.dtype
        )
        self.hessian_initialized = False
        self.coef_P2 = None
        self.norm_min_subgrad = None
        self.max_min_subgrad = None
        self.inner_tol = None

        self.n_line_search = None

        self.n_updated = 0
        self.n_active_rows = data.X.shape[0]
        self.old_active_set = None
        self.active_set = np.arange(self.coef.shape[0], dtype=np.int32)

        self.build_hessian_runtime = None
        self.inner_solver_runtime = None
        self.line_search_runtime = None
        self.quadratic_update_runtime = None

    def _record_iteration(self):
        self.n_iter += 1

        self.iteration_runtime = time.time() - self.iteration_start
        self.iteration_start = time.time()

        coef_l1 = np.sum(np.abs(self.coef))
        coef_l2 = np.linalg.norm(self.coef)
        step_l2 = np.linalg.norm(self.step)
        self.diagnostics.append(
            {
                "n_iter": self.n_iter,
                "convergence": self.norm_min_subgrad,
                "objective_fct": self.obj_val,
                "L1(coef)": coef_l1,
                "L2(coef)": coef_l2,
                "L2(step)": step_l2,
                "first_coef": self.coef[0],
                "n_coef_updated": self.n_updated,
                "n_active_cols": self.active_set.shape[0],
                "n_active_rows": self.n_active_rows,
                "n_cycles": self.n_cycles,
                "n_line_search": self.n_line_search,
                "iteration_runtime": self.iteration_runtime,
                "build_hessian_runtime": self.build_hessian_runtime,
                "inner_solver_runtime": self.inner_solver_runtime,
                "line_search_runtime": self.line_search_runtime,
                "quadratic_update_runtime": self.quadratic_update_runtime,
            }
        )


def check_convergence(
    state: IRLSState, data: IRLSData
) -> Tuple[bool, float, float, float]:
    """Calculate parameters needed to determine whether we have converged."""
    # stopping criterion for outer loop is a mix of a subgradient tolerance
    # and a step size tolerance
    # sum_i(|minimum-norm of subgrad of F(w)_i|)
    # fp_wP2 = f'(w) + w*P2
    # Note: eta, mu and score are already updated
    # this also updates the inner tolerance for the next loop!

    # L1 norm of the minimum norm subgradient
    norm_min_subgrad, max_min_subgrad = _norm_min_subgrad(
        np.arange(state.coef.shape[0], dtype=np.int32),
        state.coef,
        -state.score,
        state.data.P1,
        state.data.intercept_offset,
        data.has_lower_bounds,
        data._lower_bounds,
        data.has_upper_bounds,
        data._upper_bounds,
    )
    gradient_converged = norm_min_subgrad < state.data.gradient_tol

    # Simple L2 step length convergence criteria
    step_size = linalg.norm(state.step)
    step_size_converged = (
        state.data.step_size_tol is not None and step_size < state.data.step_size_tol
    )

    converged = gradient_converged or step_size_converged

    # the ratio of inner tolerance to the minimum subgradient norm
    # This wasn't explored in the newGLMNET paper linked above.
    # That paper essentially uses inner_tol_ratio = 1.0, but using a slightly
    # lower value is much faster.
    # By comparison, the original GLMNET paper uses inner_tol = tol.
    # So, inner_tol_ratio < 1 is sort of a compromise between the two papers.
    # Lower values tend to increase the time spent in the inner solver while
    # reducing the time spent computing quadratic approximations. We find that
    # 0.001 is generally a good number.
    inner_tol_ratio = 0.001

    if state.data.fixed_inner_tol is None:
        # Another potential rule limits the inner tol to be no smaller than tol
        # return max(norm_min_subgrad * inner_tol_ratio, tol)
        inner_tol = max(norm_min_subgrad * inner_tol_ratio, data.gradient_tol)
    else:
        inner_tol = state.data.fixed_inner_tol[0]

    return converged, norm_min_subgrad, max_min_subgrad, inner_tol


def _update_predictions(state, data, coef, X_dot_step=None, factor=1.0):
    if X_dot_step is None:
        X_dot_step = _safe_lin_pred(data.X, coef, data.offset)
    return eta_mu_objective(
        data.family,
        data.link,
        X_dot_step,
        factor,
        coef,
        state.eta,
        data.y,
        data.sample_weight,
        data.P1,
        data.P2,
        data.intercept_offset,
    )


def eta_mu_objective(
    family,
    link,
    X_dot_step,
    factor,
    coef,
    cur_eta,
    y,
    sample_weight,
    P1,
    P2,
    intercept_offset,
):
    """Calculate eta, mu, and the objective value."""
    eta, mu, deviance = family.eta_mu_deviance(
        link, factor, cur_eta, X_dot_step, y, sample_weight
    )
    obj_val = 0.5 * deviance
    obj_val += linalg.norm(P1 * coef[intercept_offset:], ord=1)
    coef_P2 = _make_coef_P2(intercept_offset, P2, coef)
    obj_val += 0.5 * (coef_P2 @ coef)
    return eta, mu, obj_val, coef_P2


@timeit("quadratic_update_runtime")
def update_quadratic(
    state: IRLSState, data: IRLSData, coef_P2
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Update the quadratic approximation."""
    gradient_rows, hessian_rows = data.family.rowwise_gradient_hessian(
        data.link,
        coef=state.coef,
        dispersion=1,
        X=data.X,
        y=data.y,
        sample_weight=data.sample_weight,
        offset=data.offset,
        eta=state.eta,
        mu=state.mu,
    )

    grad = gradient_rows @ data.X
    if data.fit_intercept:
        grad = np.concatenate(([gradient_rows.sum()], grad))
    grad -= coef_P2
    return gradient_rows, grad, hessian_rows


def _make_coef_P2(intercept_offset, P2, coef):
    out = np.empty_like(coef)

    if intercept_offset == 1:
        out[0] = 0

    C = coef[intercept_offset:]
    if P2.ndim == 1:
        out[intercept_offset:] = C * P2
    else:
        out[intercept_offset:] = C @ P2

    return out


def identify_active_set(state: IRLSState, data: IRLSData):
    """
    Find the coefficients in the active set for this iteration.

    This criteria is from section 5.3 of:
    An Improved GLMNET for L1-regularized LogisticRegression.
    Yuan, Ho, Lin. 2012
    https://www.csie.ntu.edu.tw/~cjlin/papers/l1_glmnet/long-glmnet.pdf

    Parameters
    ----------
    state
    data

    Returns
    -------
    np.ndarray
    """
    T = data.P1 - state.max_min_subgrad
    abs_score = np.abs(state.score[data.intercept_offset :])
    active = np.logical_or(state.coef[data.intercept_offset :] != 0, abs_score >= T)

    active_set = np.concatenate(
        ([0] if data.fit_intercept else [], np.where(active)[0] + data.intercept_offset)
    ).astype(np.int32)

    return active_set


@timeit("line_search_runtime")
def line_search(state: IRLSState, data: IRLSData, d: np.ndarray):
    """
    Run a backtracking line search.

    Parameters
    ----------
    state
    data
    d

    Returns
    -------
    tuple
    """
    # line search parameters
    (beta, sigma) = (0.5, 0.0001)

    # line search by sequence beta^k, k=0, 1, ..
    # F(w + lambda d) - F(w) <= lambda * bound
    # bound = sigma * (f'(w)*d + w*P2*d
    #                  +||P1 (w+d)||_1 - ||P1 w||_1)
    # This is a standard Armijo-Goldstein backtracking line search algorithm:
    # https://en.wikipedia.org/wiki/Backtracking_line_search
    P1w_1 = linalg.norm(data.P1 * state.coef[data.intercept_offset :], ord=1)
    P1wd_1 = linalg.norm(data.P1 * (state.coef + d)[data.intercept_offset :], ord=1)

    # Note: the L2 penalty term is included in the score.
    bound = sigma * (-(state.score @ d) + P1wd_1 - P1w_1)

    # The step direction in row space. We'll be multiplying this by varying
    # step sizes during the line search. Factoring this matrix-vector product
    # out of the inner loop improve performance a lot!
    X_dot_d = _safe_lin_pred(data.X, d)

    # Try progressively shorter line search steps.
    # variables suffixed with wd are for the new coefficient values
    factor = 1.0
    for _k in range(20):
        step = factor * d
        coef_wd = state.coef + step
        eta_wd, mu_wd, obj_val_wd, coef_wd_P2 = _update_predictions(
            state, data, coef_wd, X_dot_d, factor=factor
        )
        if (mu_wd.max() < 1e43) and (obj_val_wd - state.obj_val <= factor * bound):
            break
        factor *= beta
    else:
        warnings.warn(
            "Line search failed. Next iteration will be very close to current "
            "iteration. Might result in more convergence issues.",
            ConvergenceWarning,
        )

    state.n_line_search = _k

    # obj_val in the next iteration will be equal to obj_val_wd this iteration.
    # We can avoid a matrix-vector product inside _eta_mu_score_hessian by
    # returning the new eta and mu calculated here.
    # NOTE: This might accumulate some numerical error over a sufficient number
    # of iterations since we aren't calculating eta or mu from scratch but
    # instead adding the delta from the previous iteration. Maybe we should
    # completely recompute eta every N iterations?
    return state.coef + step, step, eta_wd, mu_wd, obj_val_wd, coef_wd_P2


def _get_obj_and_derivative(
    coef,
    X,
    y: np.ndarray,
    sample_weight: np.ndarray,
    P2: Union[np.ndarray, sparse.spmatrix],
    family: ExponentialDispersionModel,
    link: Link,
    offset: np.ndarray = None,
):
    mu, devp = family._mu_deviance_derivative(coef, X, y, sample_weight, link, offset)
    dev = family.deviance(y, mu, sample_weight)
    intercept = coef.size == X.shape[1] + 1
    idx = 1 if intercept else 0  # offset if coef[0] is intercept
    if P2.ndim == 1:
        L2 = P2 * coef[idx:]
    else:
        L2 = P2 @ coef[idx:]
    obj = 0.5 * dev + 0.5 * (coef[idx:] @ L2)
    objp = 0.5 * devp
    objp[idx:] += L2
    return obj, objp


def _lbfgs_solver(
    coef,
    X,
    y: np.ndarray,
    sample_weight: np.ndarray,
    P2: Union[np.ndarray, sparse.spmatrix],
    verbose: bool,
    family: ExponentialDispersionModel,
    link: Link,
    max_iter: int = 100,
    tol: float = 1e-4,
    offset: np.ndarray = None,
):
    func = functools.partial(
        _get_obj_and_derivative,
        X=X,
        y=y,
        sample_weight=sample_weight,
        P2=P2,
        family=family,
        link=link,
        offset=offset,
    )

    coef, loss, info = fmin_l_bfgs_b(
        func,
        coef,
        fprime=None,
        iprint=(verbose > 0) - 1,
        pgtol=tol,
        maxiter=max_iter,
        factr=1e2,
    )
    if info["warnflag"] == 1:
        warnings.warn(
            "lbfgs failed to converge." " Increase the number of iterations.",
            ConvergenceWarning,
        )
    elif info["warnflag"] == 2:
        warnings.warn("lbfgs failed for the reason: {}".format(info["task"]))
    n_iter_ = info["nit"]

    return coef, n_iter_, -1, None


def _trust_constr_solver(
    coef,
    X,
    y: np.ndarray,
    sample_weight: np.ndarray,
    P2: Union[np.ndarray, sparse.spmatrix],
    fit_intercept: bool,
    verbose: bool,
    family: ExponentialDispersionModel,
    link: Link,
    max_iter: int = 100,
    xtol: Optional[float] = 1e-8,
    gtol: Optional[float] = 1e-8,
    offset: np.ndarray = None,
    A_ineq: Optional[np.ndarray] = None,
    b_ineq: Optional[np.ndarray] = None,
):
    fun = functools.partial(
        _get_obj_and_derivative,
        X=X,
        y=y,
        sample_weight=sample_weight,
        P2=P2,
        family=family,
        link=link,
        offset=offset,
    )

    if (A_ineq is not None) and (b_ineq is not None):

        if fit_intercept:
            # add one column of 0's from the left
            # the intercept will not be constrained
            A_ineq_intercept = np.zeros(shape=(A_ineq.shape[0], 1))
            A_ineq_ = np.concatenate((A_ineq_intercept, A_ineq), axis=1)
        else:
            A_ineq_ = A_ineq

        # we express constraints in the form A theta <= b
        constraints = LinearConstraint(
            A=A_ineq_,
            lb=-np.Inf,
            ub=b_ineq,
        )
    else:
        constraints = ()

    res = minimize(
        fun=fun,
        x0=coef,
        jac=True,
        method="trust-constr",
        hess="2-point",
        constraints=constraints,
        options={
            "xtol": xtol,
            "gtol": gtol,
            "maxiter": max_iter,
            "verbose": 2 if verbose else 0,
        },
    )

    if not res["success"]:
        warnings.warn(
            f"trust-constr failed with message: {res['message']}",
            ConvergenceWarning,
        )

    return res["x"], res["nit"], -1, None



from abc import ABCMeta, abstractmethod
from functools import partial
from typing import Tuple, Union

import numexpr
import numpy as np
from scipy import sparse, special
from tabmat import MatrixBase, StandardizedMatrix

from ._functions import (
    binomial_logit_eta_mu_deviance,
    binomial_logit_rowwise_gradient_hessian,
    gamma_deviance,
    gamma_log_eta_mu_deviance,
    gamma_log_likelihood,
    gamma_log_rowwise_gradient_hessian,
    normal_deviance,
    normal_identity_eta_mu_deviance,
    normal_identity_rowwise_gradient_hessian,
    normal_log_likelihood,
    poisson_deviance,
    poisson_log_eta_mu_deviance,
    poisson_log_likelihood,
    poisson_log_rowwise_gradient_hessian,
    tweedie_deviance,
    tweedie_log_eta_mu_deviance,
    tweedie_log_likelihood,
    tweedie_log_rowwise_gradient_hessian,
)
from ._link import IdentityLink, Link, LogitLink, LogLink
from ._util import _safe_lin_pred, _safe_sandwich_dot


class ExponentialDispersionModel(metaclass=ABCMeta):
    r"""Base class for reproductive Exponential Dispersion Models (EDM).

    The PDF of :math:`Y \sim \mathrm{EDM}(\mu, \phi)` is given by

    .. math::

        p(y \mid \theta, \phi)
        &= c(y, \phi) \exp((\theta y - A(\theta)_ / \phi) \\
        &= \tilde{c}(y, \phi) \exp(-d(y, \mu) / (2\phi))

    with mean :math:`\mathrm{E}(Y) = A'(\theta) = \mu`, variance
    :math:`\mathrm{var}(Y) = \phi \cdot v(\mu)`, unit variance
    :math:`v(\mu)` and unit deviance :math:`d(y, \mu)`.

    Properties
    ----------
    lower_bound
    upper_bound
    include_lower_bound
    include_upper_bound

    Methods
    -------
    in_y_range
    unit_variance
    unit_variance_derivative
    variance
    variance_derivative
    unit_deviance
    unit_deviance_derivative
    deviance
    deviance_derivative
    starting_mu

    _mu_deviance_derivative
    eta_mu_deviance
    gradient_hessian

    References
    ----------
    https://en.wikipedia.org/wiki/Exponential_dispersion_model.
    """

    @property
    @abstractmethod
    def lower_bound(self) -> float:
        """Get the lower bound of values for the EDM."""
        pass

    @property
    @abstractmethod
    def upper_bound(self) -> float:
        """Get the upper bound of values for the EDM."""
        pass

    @property
    def include_lower_bound(self) -> bool:
        """Return whether ``lower_bound`` is allowed as a value of ``y``."""
        pass

    @property
    def include_upper_bound(self) -> bool:
        """Return whether ``upper_bound`` is allowed as a value of ``y``."""
        pass

    def in_y_range(self, x) -> np.ndarray:
        """Return ``True`` if ``x`` is in the valid range of the EDM.

        Parameters
        ----------
        x : array-like, shape (n_samples,)
            Target values.

        Returns
        -------
        np.ndarray
        """
        if self.include_lower_bound:
            if self.include_upper_bound:
                return np.logical_and(
                    np.greater_equal(x, self.lower_bound),
                    np.less_equal(x, self.upper_bound),
                )
            else:
                return np.logical_and(
                    np.greater_equal(x, self.lower_bound), np.less(x, self.upper_bound)
                )
        else:
            if self.include_upper_bound:
                return np.logical_and(
                    np.greater(x, self.lower_bound), np.less_equal(x, self.upper_bound)
                )
            else:
                return np.logical_and(
                    np.greater(x, self.lower_bound), np.less(x, self.upper_bound)
                )

    @abstractmethod
    def unit_variance(self, mu):
        r"""Compute the unit variance function.

        The unit variance :math:`v(\mu)` determines the variance as a function
        of the mean :math:`\mu` by
        :math:`\mathrm{var}(y_i) = (\phi / s_i) \times v(\mu_i)`. It can
        also be derived from the unit deviance :math:`d(y, \mu)` as

        .. math::

            v(\mu) = \frac{2}{\frac{\partial^2 d(y, \mu)}{\partial\mu^2}}\big|_{y=\mu}.

        See also :func:`variance`.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Predicted mean.
        """
        pass

    @abstractmethod
    def unit_variance_derivative(self, mu):
        r"""Compute the derivative of the unit variance with respect to ``mu``.

        Return :math:`v'(\mu)`.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Predicted mean.
        """
        pass

    def variance(self, mu: np.ndarray, dispersion=1, sample_weight=1) -> np.ndarray:
        r"""Compute the variance function.

        The variance of :math:`Y_i \sim \mathrm{EDM}(\mu_i, \phi / s_i)` is
        :math:`\mathrm{var}(Y_i) = (\phi / s_i) * v(\mu_i)`, with unit variance
        :math:`v(\mu)` and weights :math:`s_i`.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Predicted mean.

        dispersion : float, optional (default=1)
            Dispersion parameter :math:`\phi`.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Weights or exposure to which variance is inverse proportional.

        Returns
        -------
        array-like, shape (n_samples,)
        """
        return self.unit_variance(mu) * dispersion / sample_weight

    def variance_derivative(self, mu, dispersion=1, sample_weight=1):
        r"""Compute the derivative of the variance with respect to ``mu``.

        The derivative of the variance is equal to
        :math:`(\phi / s_i) * v'(\mu_i)`, where :math:`v(\mu)` is the unit
        variance and :math:`s_i` are weights.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Predicted mean.

        dispersion : float, optional (default=1)
            Dispersion parameter :math:`\phi`.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Weights or exposure to which variance is inverse proportional.

        Returns
        -------
        array-like, shape (n_samples,)
        """
        return self.unit_variance_derivative(mu) * dispersion / sample_weight

    @abstractmethod
    def unit_deviance(self, y, mu):
        r"""Compute the unit deviance.

        In terms of the log likelihood :math:`L`, the unit deviance is
        :math:`-2\phi\times [L(y, \mu, \phi) - L(y, y, \phi)].`

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.
        """
        pass

    def unit_deviance_derivative(self, y, mu):
        r"""Compute the derivative of the unit deviance with respect to ``mu``.

        The derivative of the unit deviance is given by
        :math:`-2 \times (y - \mu) / v(\mu)`, where :math:`v(\mu)` is the unit
        variance.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        Returns
        -------
        array-like, shape (n_samples,)
        """
        return -2 * (y - mu) / self.unit_variance(mu)

    def deviance(self, y, mu, sample_weight=1):
        r"""Compute the deviance.

        The deviance is a weighted sum of the unit deviances,
        :math:`\sum_i s_i \times d(y_i, \mu_i)`, where :math:`d(y, \mu)` is the
        unit deviance and :math:`s` are weights. In terms of the log likelihood,
        it is :math:`-2\phi \times [L(y, \mu, \phi / s) - L(y, y, \phi / s)]`.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Weights or exposure to which variance is inversely proportional.

        Returns
        -------
        float
        """
        if sample_weight is None:
            return np.sum(self.unit_deviance(y, mu))
        else:
            return np.sum(self.unit_deviance(y, mu) * sample_weight)

    def deviance_derivative(self, y, mu, sample_weight=1):
        r"""Compute the derivative of the deviance with respect to ``mu``.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,) (default=1)
            Weights or exposure to which variance is inverse proportional.

        Returns
        -------
        array-like, shape (n_samples,)
        """
        return sample_weight * self.unit_deviance_derivative(y, mu)

    def _mu_deviance_derivative(
        self,
        coef: np.ndarray,
        X,
        y: np.ndarray,
        sample_weight: np.ndarray,
        link: Link,
        offset: np.ndarray = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Compute ``mu`` and the derivative of the deviance \
            with respect to coefficients."""
        lin_pred = _safe_lin_pred(X, coef, offset)
        mu = link.inverse(lin_pred)
        d1 = link.inverse_derivative(lin_pred)
        temp = d1 * self.deviance_derivative(y, mu, sample_weight)
        if coef.size == X.shape[1] + 1:
            devp = np.concatenate(([temp.sum()], temp @ X))
        else:
            devp = temp @ X  # same as X.T @ temp
        return mu, devp

    def eta_mu_deviance(
        self,
        link: Link,
        factor: float,
        cur_eta: np.ndarray,
        X_dot_d: np.ndarray,
        y: np.ndarray,
        sample_weight: np.ndarray,
    ):
        """
        Compute ``eta``, ``mu`` and the deviance.

        Compute:
        * the linear predictor, ``eta``, as ``cur_eta + factor * X_dot_d``;
        * the link-function-transformed prediction, ``mu``;
        * the deviance.

        Returns
        -------
        numpy.ndarray, shape (X.shape[0],)
            The linear predictor, ``eta``.
        numpy.ndarray, shape (X.shape[0],)
            The link-function-transformed prediction, ``mu``.
        float
            The deviance.
        """
        # eta_out and mu_out are filled inside self._eta_mu_deviance,
        # avoiding allocating new arrays for every line search loop
        eta_out = np.empty_like(cur_eta)
        mu_out = np.empty_like(cur_eta)
        deviance = self._eta_mu_deviance(
            link, factor, cur_eta, X_dot_d, y, sample_weight, eta_out, mu_out
        )
        return eta_out, mu_out, deviance

    def _eta_mu_deviance(
        self,
        link: Link,
        factor: float,
        cur_eta: np.ndarray,
        X_dot_d: np.ndarray,
        y: np.ndarray,
        sample_weight: np.ndarray,
        eta_out: np.ndarray,
        mu_out: np.ndarray,
    ):
        """
        Update ``eta`` and ``mu`` and compute the deviance.

        This is a default implementation that should work for all valid
        distributions and link functions. To implement a custom optimized
        version for a specific distribution and link function, please override
        this function in the subclass.

        Returns
        -------
        float
        """
        eta_out[:] = cur_eta + factor * X_dot_d
        mu_out[:] = link.inverse(eta_out)
        return self.deviance(y, mu_out, sample_weight=sample_weight)

    def rowwise_gradient_hessian(
        self,
        link: Link,
        coef: np.ndarray,
        dispersion,
        X: Union[MatrixBase, StandardizedMatrix],
        y: np.ndarray,
        sample_weight: np.ndarray,
        eta: np.ndarray,
        mu: np.ndarray,
        offset: np.ndarray = None,
    ):
        """
        Compute the gradient and negative Hessian of the log likelihood row-wise.

        Returns
        -------
        numpy.ndarray, shape (X.shape[0],)
            The gradient of the log likelihood, row-wise.
        numpy.ndarray, shape (X.shape[0],)
            The negative Hessian of the log likelihood, row-wise.
        """
        gradient_rows = np.empty_like(mu)
        hessian_rows = np.empty_like(mu)
        self._rowwise_gradient_hessian(
            link, y, sample_weight, eta, mu, gradient_rows, hessian_rows
        )

        # To form the full Hessian matrix from the IRLS sample_weight:
        # hessian_matrix = _safe_sandwich_dot(X, hessian_rows, intercept=intercept)
        return gradient_rows, hessian_rows

    def _rowwise_gradient_hessian(
        self, link, y, sample_weight, eta, mu, gradient_rows, hessian_rows
    ):
        """
        Update ``gradient_rows`` and ``hessian_rows`` in place.

        This is a default implementation that should work for all valid
        distributions and link functions. To implement a custom optimized
        version for a specific distribution and link function, please override
        this function in the subclass.
        """
        # FOR TWEEDIE: sigma_inv = weights / (mu ** p) during optimization bc phi = 1
        sigma_inv = get_one_over_variance(self, link, mu, eta, 1.0, sample_weight)
        d1 = link.inverse_derivative(eta)  # = h'(eta)
        # Alternatively:
        # h'(eta) = h'(g(mu)) = 1/g'(mu), note that h is inverse of g
        # d1 = 1./link.derivative(mu)
        d1_sigma_inv = d1 * sigma_inv
        gradient_rows[:] = d1_sigma_inv * (y - mu)
        hessian_rows[:] = d1 * d1_sigma_inv

    def _fisher_information(
        self, link, X, y, mu, sample_weight, dispersion, fit_intercept
    ):
        """Compute the expected information matrix.

        Parameters
        ----------
        link : Link
            A link function (i.e. an instance of :class:`~glum._link.Link`).
        X : array-like
            Training data.
        y : array-like
            Target values.
        mu : array-like
            Predicted mean.
        sample_weight : array-like
            Weights or exposure to which variance is inversely proportional.
        dispersion : float
            The dispersion parameter.
        fit_intercept : bool
            Whether the model has an intercept.
        """
        W = (link.inverse_derivative(link.link(mu)) ** 2) * get_one_over_variance(
            self, link, mu, link.inverse(mu), dispersion, sample_weight
        )

        return _safe_sandwich_dot(X, W, intercept=fit_intercept)

    def _observed_information(
        self, link, X, y, mu, sample_weight, dispersion, fit_intercept
    ):
        """Compute the observed information matrix.

        Parameters
        ----------
        X : array-like
            Training data.
        y : array-like
            Target values.
        mu : array-like
            Predicted mean.
        sample_weight : array-like
            Weights or exposure to which variance is inversely proportional.
        dispersion : float
            The dispersion parameter.
        fit_intercept : bool
            Whether the model has an intercept.
        """
        linpred = link.link(mu)

        W = (
            -link.inverse_derivative2(linpred) * (y - mu)
            + (link.inverse_derivative(linpred) ** 2)
            * (
                1
                + (y - mu) * self.unit_variance_derivative(mu) / self.unit_variance(mu)
            )
        ) * get_one_over_variance(self, link, mu, linpred, dispersion, sample_weight)

        return _safe_sandwich_dot(X, W, intercept=fit_intercept)

    def _score_matrix(self, link, X, y, mu, sample_weight, dispersion, fit_intercept):
        """Compute the score.

        Parameters
        ----------
        X : array-like
            Training data.
        y : array-like
            Target values.
        mu : array-like
            Predicted mean.
        sample_weight : array-like
            Weights or exposure to which variance is inversely proportional.
        dispersion : float
            The dispersion parameter.
        fit_intercept : bool
            Whether the model has an intercept.
        """
        linpred = link.link(mu)

        W = (
            get_one_over_variance(self, link, mu, linpred, dispersion, sample_weight)
            * link.inverse_derivative(linpred)
            * (y - mu)
        ).reshape(-1, 1)

        if fit_intercept:
            if sparse.issparse(X):
                return sparse.hstack((W, X.multiply(W)))
            else:
                return np.hstack((W, np.multiply(X, W)))
        else:
            if sparse.issparse(X):
                return X.multiply(W)
            else:
                return np.multiply(X, W)

    def dispersion(self, y, mu, sample_weight=None, ddof=1, method="pearson") -> float:
        r"""Estimate the dispersion parameter :math:`\phi`.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Weights or exposure to which variance is inversely proportional.

        ddof : int, optional (default=1)
            Degrees of freedom consumed by the model for ``mu``.

        method = {'pearson', 'deviance'}, optional (default='pearson')
            Whether to base the estimate on the Pearson residuals or the deviance.

        Returns
        -------
        float
        """
        y, mu, sample_weight = _as_float_arrays(y, mu, sample_weight)

        if method == "pearson":
            pearson_residuals = ((y - mu) ** 2) / self.unit_variance(mu)
            if sample_weight is None:
                numerator = pearson_residuals.sum()
            else:
                numerator = np.dot(pearson_residuals, sample_weight)
        elif method == "deviance":
            numerator = self.deviance(y, mu, sample_weight)
        else:
            raise NotImplementedError(f"Method {method} hasn't been implemented.")

        if sample_weight is None:
            return numerator / (len(y) - ddof)
        else:
            return numerator / (sample_weight.sum() - ddof)


class TweedieDistribution(ExponentialDispersionModel):
    r"""A class for the Tweedie distribution.

    A Tweedie distribution with mean :math:`\mu = \mathrm{E}(Y)` is uniquely
    defined by its mean-variance relationship
    :math:`\mathrm{var}(Y) \propto \mu^{\mathrm{power}}`.

    Special cases are:

    ====== ================
    Power  Distribution
    ====== ================
    0      Normal
    1      Poisson
    (1, 2) Compound Poisson
    2      Gamma
    3      Inverse Gaussian
    ====== ================

    Parameters
    ----------
    power : float, optional (default=0)
        The variance power of the `unit_variance`
        :math:`v(\mu) = \mu^{\mathrm{power}}`. For
        :math:`0 < \mathrm{power} < 1`, no distribution exists.
    """

    upper_bound = np.Inf
    include_upper_bound = False

    def __init__(self, power=0):
        # validate power and set _upper_bound, _include_upper_bound attrs
        self.power = power

    @property
    def lower_bound(self) -> Union[float, int]:
        """Return the lowest value of ``y`` allowed."""
        if self.power <= 0:
            return -np.Inf
        if self.power >= 1:
            return 0
        raise ValueError

    @property
    def include_lower_bound(self) -> bool:
        """Return whether ``lower_bound`` is allowed as a value of ``y``."""
        if self.power <= 0:
            return False
        if (self.power >= 1) and (self.power < 2):
            return True
        if self.power >= 2:
            return False
        raise ValueError

    @property
    def power(self) -> float:
        """Return the Tweedie power parameter."""
        return self._power

    @power.setter
    def power(self, power):

        if not isinstance(power, (int, float)):
            raise TypeError(f"power must be an int or float, input was {power}")
        if (power > 0) and (power < 1):
            raise ValueError("For 0<power<1, no distribution exists.")

        # Prevents upcasting when working with 32-bit data
        self._power = power if isinstance(power, int) else np.float32(power)

    def unit_variance(self, mu: np.ndarray) -> np.ndarray:
        """Compute the unit variance of a Tweedie distribution ``v(mu) = mu^power``.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Predicted mean.

        Returns
        -------
        numpy.ndarray, shape (n_samples,)
        """
        p = self.power  # noqa: F841
        return numexpr.evaluate("mu ** p")

    def unit_variance_derivative(self, mu: np.ndarray) -> np.ndarray:
        r"""Compute the derivative of the unit variance of a Tweedie distribution.

        Equation: :math:`v(\mu) = p \times \mu^{(p-1)}`.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Predicted mean.

        Returns
        -------
        numpy.ndarray, shape (n_samples,)
        """
        p = self.power  # noqa: F841
        return numexpr.evaluate("p * mu ** (p - 1)")

    def deviance(self, y, mu, sample_weight=None) -> float:
        """Compute the deviance.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Sample weights.
        """
        p = self.power
        y, mu, sample_weight = _as_float_arrays(y, mu, sample_weight)
        sample_weight = np.ones_like(y) if sample_weight is None else sample_weight

        # NOTE: the dispersion parameter is only necessary to convey
        # type information on account of a bug in Cython

        if p == 0:
            return normal_deviance(y, sample_weight, mu, dispersion=1.0)
        if p == 1:
            return poisson_deviance(y, sample_weight, mu, dispersion=1.0)
        elif p == 2:
            return gamma_deviance(y, sample_weight, mu, dispersion=1.0)
        else:
            return tweedie_deviance(y, sample_weight, mu, p=float(p))

    def unit_deviance(self, y, mu):
        """Get the deviance of each observation."""
        p = self.power
        if p == 0:  # Normal distribution
            return (y - mu) ** 2
        if p == 1:  # Poisson distribution
            return 2 * (special.xlogy(y, y / mu) - y + mu)
        elif p == 2:  # Gamma distribution
            return 2 * (np.log(mu / y) + y / mu - 1)
        else:
            mu1mp = mu ** (1 - p)
            return 2 * (
                (np.maximum(y, 0) ** (2 - p)) / ((1 - p) * (2 - p))
                - y * mu1mp / (1 - p)
                + mu * mu1mp / (2 - p)
            )

    def _rowwise_gradient_hessian(
        self, link, y, sample_weight, eta, mu, gradient_rows, hessian_rows
    ):
        f = None
        if self.power == 0 and isinstance(link, IdentityLink):
            f = normal_identity_rowwise_gradient_hessian
        elif self.power == 1 and isinstance(link, LogLink):
            f = poisson_log_rowwise_gradient_hessian
        elif self.power == 2 and isinstance(link, LogLink):
            f = gamma_log_rowwise_gradient_hessian
        elif 1 < self.power < 2 and isinstance(link, LogLink):
            f = partial(tweedie_log_rowwise_gradient_hessian, p=self.power)

        if f is not None:
            return f(y, sample_weight, eta, mu, gradient_rows, hessian_rows)

        return super()._rowwise_gradient_hessian(
            link, y, sample_weight, eta, mu, gradient_rows, hessian_rows
        )

    def _eta_mu_deviance(
        self,
        link: Link,
        factor: float,
        cur_eta: np.ndarray,
        X_dot_d: np.ndarray,
        y: np.ndarray,
        sample_weight: np.ndarray,
        eta_out: np.ndarray,
        mu_out: np.ndarray,
    ):
        f = None
        if self.power == 0 and isinstance(link, IdentityLink):
            f = normal_identity_eta_mu_deviance
        elif self.power == 1 and isinstance(link, LogLink):
            f = poisson_log_eta_mu_deviance
        elif self.power == 2 and isinstance(link, LogLink):
            f = gamma_log_eta_mu_deviance
        elif 1 < self.power < 2 and isinstance(link, LogLink):
            f = partial(tweedie_log_eta_mu_deviance, p=self.power)

        if f is not None:
            return f(cur_eta, X_dot_d, y, sample_weight, eta_out, mu_out, factor)

        return super()._eta_mu_deviance(
            link, factor, cur_eta, X_dot_d, y, sample_weight, eta_out, mu_out
        )

    def log_likelihood(self, y, mu, sample_weight=None, dispersion=None) -> float:
        r"""Compute the log likelihood.

        For ``1 < power < 2``, we use the series approximation by Dunn and Smyth
        (2005) to compute the normalization term.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Sample weights.

        dispersion : float, optional (default=None)
            Dispersion parameter :math:`\phi`. Estimated if ``None``.
        """
        p = self.power
        y, mu, sample_weight = _as_float_arrays(y, mu, sample_weight)
        sample_weight = np.ones_like(y) if sample_weight is None else sample_weight

        if (p != 1) and (dispersion is None):
            dispersion = self.dispersion(y, mu, sample_weight)

        if p == 0:
            return normal_log_likelihood(y, sample_weight, mu, float(dispersion))
        if p == 1:
            # NOTE: the dispersion parameter is only necessary to convey
            # type information on account of a bug in Cython
            return poisson_log_likelihood(y, sample_weight, mu, 1.0)
        elif p == 2:
            return gamma_log_likelihood(y, sample_weight, mu, float(dispersion))
        elif p < 2:
            return tweedie_log_likelihood(
                y, sample_weight, mu, float(p), float(dispersion)
            )
        else:
            raise NotImplementedError

    def dispersion(self, y, mu, sample_weight=None, ddof=1, method="pearson") -> float:
        r"""Estimate the dispersion parameter :math:`\phi`.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=None)
            Weights or exposure to which variance is inversely proportional.

        ddof : int, optional (default=1)
            Degrees of freedom consumed by the model for ``mu``.

        method = {'pearson', 'deviance'}, optional (default='pearson')
            Whether to base the estimate on the Pearson residuals or the deviance.

        Returns
        -------
        float
        """
        p = self.power  # noqa: F841
        y, mu, sample_weight = _as_float_arrays(y, mu, sample_weight)

        if method == "pearson":
            formula = "((y - mu) ** 2) / (mu ** p)"
            if sample_weight is None:
                return numexpr.evaluate(formula).sum() / (len(y) - ddof)
            else:
                formula = f"sample_weight * {formula}"
                return numexpr.evaluate(formula).sum() / (sample_weight.sum() - ddof)

        return super().dispersion(
            y, mu, sample_weight=sample_weight, ddof=ddof, method=method
        )


class NormalDistribution(TweedieDistribution):
    """Class for the Normal (a.k.a. Gaussian) distribution."""

    def __init__(self):
        super().__init__(power=0)


class PoissonDistribution(TweedieDistribution):
    """Class for the scaled Poisson distribution."""

    def __init__(self):
        super().__init__(power=1)


class GammaDistribution(TweedieDistribution):
    """Class for the Gamma distribution."""

    def __init__(self):
        super().__init__(power=2)


class InverseGaussianDistribution(TweedieDistribution):
    """Class for the scaled Inverse Gaussian distribution."""

    def __init__(self):
        super().__init__(power=3)


class GeneralizedHyperbolicSecant(ExponentialDispersionModel):
    """A class for the Generalized Hyperbolic Secant (GHS) distribution.

    The GHS distribution is for targets ``y`` in ``(-∞, +∞)``.
    """

    lower_bound = -np.Inf
    upper_bound = np.Inf
    include_lower_bound = False
    include_upper_bound = False

    def unit_variance(self, mu: np.ndarray) -> np.ndarray:
        """Get the unit-level expected variance.

        See superclass documentation.

        Parameters
        ----------
        mu : array-like or float

        Returns
        -------
        array-like
        """
        return 1 + mu ** 2

    def unit_variance_derivative(self, mu: np.ndarray) -> np.ndarray:
        """Get the derivative of the unit variance.

        See superclass documentation.

        Parameters
        ----------
        mu : array-like or float

        Returns
        -------
        array-like
        """
        return 2 * mu

    def unit_deviance(self, y: np.ndarray, mu: np.ndarray) -> np.ndarray:
        """Get the unit-level deviance.

        See superclass documentation.

        Parameters
        ----------
        y : array-like
        mu : array-like

        Returns
        -------
        array-like
        """
        return 2 * y * (np.arctan(y) - np.arctan(mu)) + np.log(
            (1 + mu ** 2) / (1 + y ** 2)
        )


class BinomialDistribution(ExponentialDispersionModel):
    """A class for the Binomial distribution.

    The Binomial distribution is for targets ``y`` in ``[0, 1]``.
    """

    lower_bound = 0
    upper_bound = 1
    include_lower_bound = True
    include_upper_bound = True

    def __init__(self):
        return

    def unit_variance(self, mu: np.ndarray) -> np.ndarray:
        """Get the unit-level expected variance.

        See superclass documentation.

        Parameters
        ----------
        mu : array-like

        Returns
        -------
        array-like
        """
        return mu * (1 - mu)

    def unit_variance_derivative(self, mu):
        """Get the derivative of the unit variance.

        See superclass documentation.

        Parameters
        ----------
        mu : array-like or float

        Returns
        -------
        array-like
        """
        return 1 - 2 * mu

    def unit_deviance(self, y: np.ndarray, mu: np.ndarray) -> np.ndarray:
        """Get the unit-level deviance.

        See superclass documentation.

        Parameters
        ----------
        y : array-like
        mu : array-like

        Returns
        -------
        array-like
        """
        # see Wooldridge and Papke (1996) for the fractional case
        return -2 * (special.xlogy(y, mu) + special.xlogy(1 - y, 1 - mu))

    def _rowwise_gradient_hessian(
        self, link, y, sample_weight, eta, mu, gradient_rows, hessian_rows
    ):
        if isinstance(link, LogitLink):
            return binomial_logit_rowwise_gradient_hessian(
                y, sample_weight, eta, mu, gradient_rows, hessian_rows
            )
        return super()._rowwise_gradient_hessian(
            link, y, sample_weight, eta, mu, gradient_rows, hessian_rows
        )

    def _eta_mu_deviance(
        self,
        link: Link,
        factor: float,
        cur_eta: np.ndarray,
        X_dot_d: np.ndarray,
        y: np.ndarray,
        sample_weight: np.ndarray,
        eta_out: np.ndarray,
        mu_out: np.ndarray,
    ):
        if isinstance(link, LogitLink):
            return binomial_logit_eta_mu_deviance(
                cur_eta, X_dot_d, y, sample_weight, eta_out, mu_out, factor
            )
        return super()._eta_mu_deviance(
            link, factor, cur_eta, X_dot_d, y, sample_weight, eta_out, mu_out
        )

    def log_likelihood(self, y, mu, sample_weight=None, dispersion=1) -> float:
        """Compute the log likelihood.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=1)
            Sample weights.

        dispersion : float, optional (default=1)
            Ignored.
        """
        ll = special.xlogy(y, mu) + special.xlogy(1 - y, 1 - mu)
        return np.sum(ll) if sample_weight is None else np.dot(ll, sample_weight)

    def dispersion(self, y, mu, sample_weight=None, ddof=1, method="pearson") -> float:
        r"""Estimate the dispersion parameter :math:`\phi`.

        Parameters
        ----------
        y : array-like, shape (n_samples,)
            Target values.

        mu : array-like, shape (n_samples,)
            Predicted mean.

        sample_weight : array-like, shape (n_samples,), optional (default=None)
            Weights or exposure to which variance is inversely proportional.

        ddof : int, optional (default=1)
            Degrees of freedom consumed by the model for ``mu``.

        method = {'pearson', 'deviance'}, optional (default='pearson')
            Whether to base the estimate on the Pearson residuals or the deviance.

        Returns
        -------
        float
        """
        y, mu, sample_weight = _as_float_arrays(y, mu, sample_weight)

        if method == "pearson":
            formula = "((y - mu) ** 2) / (mu * (1 - mu))"
            if sample_weight is None:
                return numexpr.evaluate(formula).sum() / (len(y) - ddof)
            else:
                formula = f"sample_weight * {formula}"
                return numexpr.evaluate(formula).sum() / (sample_weight.sum() - ddof)

        return super().dispersion(
            y, mu, sample_weight=sample_weight, ddof=ddof, method=method
        )


def guess_intercept(
    y: np.ndarray,
    sample_weight: np.ndarray,
    link: Link,
    distribution: ExponentialDispersionModel,
    eta: Union[np.ndarray, float] = None,
):
    """
    Say we want to find the scalar `b` that minimizes ``LL(eta + b)``, with \
    ``eta`` fixed.

    An exact solution exists for Tweedie distributions with a log link and for
    the normal distribution with identity link. An exact solution also exists
    for the case of logit with no offset.

    If the distribution and corresponding link are something else, we use the
    Tweedie or normal solution, depending on the link function.
    """
    avg_y = np.average(y, weights=sample_weight)

    if isinstance(link, IdentityLink):
        # This is only correct for normal. For other distributions, answer is unknown,
        # but assume that we want sum(y) = sum(mu)
        if eta is None:
            return avg_y
        avg_eta = eta if np.isscalar(eta) else np.average(eta, weights=sample_weight)
        return avg_y - avg_eta
    elif isinstance(link, LogLink):
        # This is only correct for Tweedie
        log_avg_y = np.log(avg_y)
        assert np.isfinite(log_avg_y).all()

        if eta is None:
            return log_avg_y
        mu = np.exp(eta)
        if isinstance(distribution, TweedieDistribution):
            p = distribution.power
        else:
            p = 1  # Like Poisson
        if np.isscalar(mu):
            first = np.log(y.dot(sample_weight) * mu ** (1 - p))
            second = np.log(sample_weight.sum() * mu ** (2 - p))
        else:
            first = np.log((y * mu ** (1 - p)).dot(sample_weight))
            second = np.log((mu ** (2 - p)).dot(sample_weight))
        return first - second
    elif isinstance(link, LogitLink):
        log_odds = np.log(avg_y) - np.log(np.average(1 - y, weights=sample_weight))
        if eta is None:
            return log_odds
        avg_eta = eta if np.isscalar(eta) else np.average(eta, weights=sample_weight)
        return log_odds - avg_eta
    else:
        return link.link(y.dot(sample_weight))


def get_one_over_variance(
    distribution: ExponentialDispersionModel,
    link: Link,
    mu: np.ndarray,
    eta: np.ndarray,
    dispersion,
    sample_weight: np.ndarray,
):
    """
    Get one over the variance.

    For Tweedie: ``sigma_inv = sample_weight / (mu ** p)`` during optimization,
    because ``phi = 1``.

    For Binomial with Logit link: Simplifies to
    ``variance = phi / ( sample_weight * (exp(eta) + 2 + exp(-eta)))``.
    More numerically accurate.
    """
    if isinstance(distribution, BinomialDistribution) and isinstance(link, LogitLink):
        max_float_for_exp = np.log(np.finfo(eta.dtype).max / 10)
        if np.any(np.abs(eta) > max_float_for_exp):
            eta = np.clip(eta, -max_float_for_exp, max_float_for_exp)  # type: ignore
        return sample_weight * (np.exp(eta) + 2 + np.exp(-eta)) / dispersion
    return 1.0 / distribution.variance(
        mu, dispersion=dispersion, sample_weight=sample_weight
    )


def _as_float_arrays(*args):
    """Convert to a float array, passing ``None`` through, and broadcast."""
    never_broadcast = {}  # type: ignore
    maybe_broadcast = {}
    always_broadcast = {}

    for ix, arg in enumerate(args):
        if isinstance(arg, (int, float)):
            maybe_broadcast[ix] = np.array([arg], dtype="float")
        elif arg is None:
            never_broadcast[ix] = None
        else:
            always_broadcast[ix] = np.asanyarray(arg, dtype="float")

    if always_broadcast and maybe_broadcast:
        to_broadcast = {**always_broadcast, **maybe_broadcast}
        _broadcast = np.broadcast_arrays(*to_broadcast.values())
        broadcast = dict(zip(to_broadcast.keys(), _broadcast))
    elif always_broadcast:
        _broadcast = np.broadcast_arrays(*always_broadcast.values())
        broadcast = dict(zip(always_broadcast.keys(), _broadcast))
    else:
        broadcast = maybe_broadcast  # possibly `{}`

    out = {**never_broadcast, **broadcast}

    return [out[ix] for ix in range(len(args))]



import logging
from typing import Union

import numpy as np
import pandas as pd
from scipy import sparse
from tabmat import MatrixBase, StandardizedMatrix

_logger = logging.getLogger(__name__)


def _asanyarray(x, **kwargs):
    """``np.asanyarray`` with passthrough for scalars."""
    return x if pd.api.types.is_scalar(x) else np.asanyarray(x, **kwargs)


def _align_df_categories(df, dtypes) -> pd.DataFrame:
    """Align data types for prediction.

    This function checks that categorical columns have same categories in the
    same order as specified in ``dtypes``. If an entry has a category not
    specified in ``dtypes``, it will be set to ``numpy.nan``.

    Parameters
    ----------
    df : pandas.DataFrame
    dtypes : Dict[str, Union[str, type, pandas.core.dtypes.base.ExtensionDtype]]
    """
    if not isinstance(df, pd.DataFrame):
        raise TypeError(f"Expected `pandas.DataFrame'; got {type(df)}.")

    changed_dtypes = {}

    categorical_dtypes = [
        column
        for column, dtype in dtypes.items()
        if pd.api.types.is_categorical_dtype(dtype) and (column in df)
    ]

    for column in categorical_dtypes:
        if not pd.api.types.is_categorical_dtype(df[column]):
            _logger.info(f"Casting {column} to categorical.")
            changed_dtypes[column] = df[column].astype(dtypes[column])
        elif list(df[column].cat.categories) != list(dtypes[column].categories):
            _logger.info(f"Aligning categories of {column}.")
            changed_dtypes[column] = df[column].cat.set_categories(
                dtypes[column].categories
            )

    if changed_dtypes:
        df = df.assign(**changed_dtypes)

    return df


def _safe_lin_pred(
    X: Union[MatrixBase, StandardizedMatrix],
    coef: np.ndarray,
    offset: np.ndarray = None,
) -> np.ndarray:
    """Compute the linear predictor taking care if intercept is present."""
    idx_offset = 0 if X.shape[1] == coef.shape[0] else 1
    nonzero_coefs = np.where(coef[idx_offset:] != 0.0)[0].astype(np.int32)
    res = X.matvec(coef[idx_offset:], cols=nonzero_coefs)

    if idx_offset == 1:
        res += coef[0]
    if offset is not None:
        return res + offset
    return res


def _safe_sandwich_dot(
    X: Union[MatrixBase, StandardizedMatrix],
    d: np.ndarray,
    rows: np.ndarray = None,
    cols: np.ndarray = None,
    intercept=False,
) -> np.ndarray:
    """
    Compute sandwich product ``X.T @ diag(d) @ X``.

    With ``intercept=True``, ``X`` is treated as if a column of 1 were appended
    as first column of ``X``. ``X`` can be sparse; ``d`` must be an ndarray.
    Always returns an ndarray.
    """
    result = X.sandwich(d, rows, cols)
    if isinstance(result, sparse.dia_matrix):
        result = result.A

    if intercept:
        dim = result.shape[0] + 1
        res_including_intercept = np.empty((dim, dim), dtype=X.dtype)
        res_including_intercept[0, 0] = d.sum()
        res_including_intercept[1:, 0] = X.transpose_matvec(d, rows, cols)
        res_including_intercept[0, 1:] = res_including_intercept[1:, 0]
        res_including_intercept[1:, 1:] = result
    else:
        res_including_intercept = result
    return res_including_intercept


def _safe_toarray(X) -> np.ndarray:
    """Return a numpy array."""
    if sparse.issparse(X):
        return X.toarray()
    else:
        return np.asarray(X)



from __future__ import division

import copy
from typing import Optional, Union

import numpy as np
from joblib import Parallel, delayed
from sklearn.model_selection._split import check_cv

from ._distribution import ExponentialDispersionModel
from ._glm import (
    ArrayLike,
    GeneralizedLinearRegressorBase,
    _standardize,
    _unstandardize,
    check_bounds,
    initialize_start_params,
    is_pos_semidef,
    setup_p1,
    setup_p2,
)
from ._link import Link, LogLink
from ._util import _safe_lin_pred


class GeneralizedLinearRegressorCV(GeneralizedLinearRegressorBase):
    """Generalized linear model with iterative fitting along a regularization path.

    The best model is selected by cross-validation.

    Cross-validated regression via a Generalized Linear Model (GLM) with
    penalties. For more on GLMs and on these parameters, see the documentation
    for :class:`GeneralizedLinearRegressor`. CV conventions follow
    :class:`sklearn.linear_model.LassoCV`.

    Parameters
    ----------
    l1_ratio : float or array of floats, optional (default=0)
        If you pass ``l1_ratio`` as an array, the ``fit`` method will choose the
        best value of ``l1_ratio`` and store it as ``self.l1_ratio``.

    P1 : {'identity', array-like}, shape (n_features,), optional (default='identity')
        With this array, you can exclude coefficients from the L1 penalty.
        Set the corresponding value to 1 (include) or 0 (exclude). The
        default value ``'identity'`` is the same as a 1d array of ones.
        Note that ``n_features = X.shape[1]``. If ``X`` is a pandas DataFrame
        with a categorical dtype and P1 has the same size as the number of columns,
        the penalty of the categorical column will be applied to all the levels of
        the categorical.

    P2 : {'identity', array-like, sparse matrix}, shape (n_features,) \
            or (n_features, n_features), optional (default='identity')
        With this option, you can set the P2 matrix in the L2 penalty
        ``w*P2*w``. This gives a fine control over this penalty (Tikhonov
        regularization). A 2d array is directly used as the square matrix P2. A
        1d array is interpreted as diagonal (square) matrix. The default
        ``'identity'`` sets the identity matrix, which gives the usual squared
        L2-norm. If you just want to exclude certain coefficients, pass a 1d
        array filled with 1 and 0 for the coefficients to be excluded. Note that
        P2 must be positive semi-definite. If ``X`` is a pandas DataFrame
        with a categorical dtype and P2 has the same size as the number of columns,
        the penalty of the categorical column will be applied to all the levels of
        the categorical. Note that if P2 is two-dimensional, its size needs to be
        of the same length as the expanded ``X`` matrix.

    fit_intercept : bool, optional (default=True)
        Specifies if a constant (a.k.a. bias or intercept) should be
        added to the linear predictor (``X * coef + intercept``).

    family : {'normal', 'poisson', 'gamma', 'inverse.gaussian', 'binomial'} \
            or ExponentialDispersionModel, optional (default='normal')
        The distributional assumption of the GLM, i.e. which distribution from
        the EDM, specifies the loss function to be minimized.

    link : {'auto', 'identity', 'log', 'logit'} or Link, optional (default='auto')
        The link function of the GLM, i.e. mapping from linear predictor
        (``X * coef``) to expectation (``mu``). Option ``'auto'`` sets the link
        depending on the chosen family as follows:

        - ``'identity'`` for family ``'normal'``
        - ``'log'`` for families ``'poisson'``, ``'gamma'`` and
          ``'inverse.gaussian'``
        - ``'logit'`` for family ``'binomial'``

    solver : {'auto', 'irls-cd', 'irls-ls', 'lbfgs'}, optional (default='auto')
        Algorithm to use in the optimization problem:

        - ``'auto'``: ``'irls-ls'`` if ``l1_ratio`` is zero and ``'irls-cd'``
          otherwise.
        - ``'irls-cd'``: Iteratively reweighted least squares with a coordinate
          descent inner solver. This can deal with L1 as well as L2 penalties.
          Note that in order to avoid unnecessary memory duplication of X in the
          ``fit`` method, ``X`` should be directly passed as a
          Fortran-contiguous Numpy array or sparse CSC matrix.
        - ``'irls-ls'``: Iteratively reweighted least squares with a least
          squares inner solver. This algorithm cannot deal with L1 penalties.
        - ``'lbfgs'``: Scipy's L-BFGS-B optimizer. It cannot deal with L1
          penalties.

    max_iter : int, optional (default=100)
        The maximal number of iterations for solver algorithms.

    gradient_tol : float, optional (default=None)
        Stopping criterion. If ``None``, solver-specific defaults will be used.
        The default value for most solvers is ``1e-4``, except for
        ``'trust-constr'``, which requires more conservative convergence
        settings and has a default value of ``1e-8``.

        For the IRLS-LS, L-BFGS and trust-constr solvers, the iteration
        will stop when ``max{|g_i|, i = 1, ..., n} <= tol``, where ``g_i`` is
        the ``i``-th component of the gradient (derivative) of the objective
        function. For the CD solver, convergence is reached when
        ``sum_i(|minimum norm of g_i|)``, where ``g_i`` is the subgradient of
        the objective and the minimum norm of ``g_i`` is the element of the
        subgradient with the smallest L2 norm.

        If you wish to only use a step-size tolerance, set ``gradient_tol``
        to a very small number.

    step_size_tol: float, optional (default=None)
        Alternative stopping criterion. For the IRLS-LS and IRLS-CD solvers, the
        iteration will stop when the L2 norm of the step size is less than
        ``step_size_tol``. This stopping criterion is disabled when
        ``step_size_tol`` is ``None``.

    hessian_approx: float, optional (default=0.0)
        The threshold below which data matrix rows will be ignored for updating
        the Hessian. See the algorithm documentation for the IRLS algorithm
        for further details.

    warm_start : bool, optional (default=False)
        Whether to reuse the solution of the previous call to ``fit``
        as initialization for ``coef_`` and ``intercept_`` (supersedes
        ``start_params``). If ``False`` or if the attribute ``coef_`` does not
        exist (first call to ``fit``), ``start_params`` sets the start values
        for ``coef_`` and ``intercept_``.

    n_alphas : int, optional (default=100)
        Number of alphas along the regularization path

    alphas : array-like, optional (default=None)
        List of alphas for which to compute the models. If ``None``, the alphas
        are set automatically. Setting ``None`` is preferred.

    min_alpha_ratio : float, optional (default=None)
        Length of the path. ``min_alpha_ratio=1e-6`` means that
        ``min_alpha / max_alpha = 1e-6``. If ``None``, ``1e-6`` is used.

    min_alpha : float, optional (default=None)
        Minimum alpha to estimate the model with. The grid will then be created
        over ``[max_alpha, min_alpha]``.

    start_params : array-like, shape (n_features*,), optional (default=None)
        Relevant only if ``warm_start`` is ``False`` or if ``fit`` is called
        for the first time (so that ``self.coef_`` does not exist yet). If
        ``None``, all coefficients are set to zero and the start value for the
        intercept is the weighted average of ``y`` (If ``fit_intercept`` is
        ``True``). If an array, used directly as start values; if
        ``fit_intercept`` is ``True``, its first element is assumed to be the
        start value for the ``intercept_``. Note that
        ``n_features* = X.shape[1] + fit_intercept``, i.e. it includes the
        intercept.

    selection : str, optional (default='cyclic')
        For the CD solver 'cd', the coordinates (features) can be updated in
        either cyclic or random order. If set to ``'random'``, a random
        coefficient is updated every iteration rather than looping over features
        sequentially in the same order, which often leads to significantly
        faster convergence, especially when ``gradient_tol`` is higher than
        ``1e-4``.

    random_state : int or RandomState, optional (default=None)
        The seed of the pseudo random number generator that selects a random
        feature to be updated for the CD solver. If an integer, ``random_state``
        is the seed used by the random number generator; if a
        :class:`RandomState` instance, ``random_state`` is the random number
        generator; if ``None``, the random number generator is the
        :class:`RandomState` instance used by ``np.random``. Used when
        ``selection`` is ``'random'``.

    copy_X : bool, optional (default=None)
        Whether to copy ``X``. Since ``X`` is never modified by
        :class:`GeneralizedLinearRegressor`, this is unlikely to be needed; this
        option exists mainly for compatibility with other scikit-learn
        estimators. If ``False``, ``X`` will not be copied and there will be an
        error if you pass an ``X`` in the wrong format, such as providing
        integer ``X`` and float ``y``. If ``None``, ``X`` will not be copied
        unless it is in the wrong format.

    check_input : bool, optional (default=True)
        Whether to bypass several checks on input: ``y`` values in range of
        ``family``, ``sample_weight`` non-negative, ``P2`` positive
        semi-definite. Don't use this parameter unless you know what you are
        doing.

    verbose : int, optional (default=0)
        For the IRLS solver, any positive number will result in a pretty
        progress bar showing convergence. This features requires having the
        tqdm package installed. For the L-BFGS solver, set ``verbose`` to any
        positive number for verbosity.

    scale_predictors: bool, optional (default=False)
        If ``True``, estimate a scaled model where all predictors have a
        standard deviation of 1. This can result in better estimates if
        predictors are on very different scales (for example, centimeters and
        kilometers).

        Advanced developer note: Internally, predictors are always rescaled for
        computational reasons, but this only affects results if
        ``scale_predictors`` is ``True``.

    lower_bounds : array-like, shape (n_features), optional (default=None)
        Set a lower bound for the coefficients. Setting bounds forces the use
        of the coordinate descent solver (``'irls-cd'``).

    upper_bounds : array-like, shape=(n_features), optional (default=None)
        See ``lower_bounds``.

    A_ineq : array-like, shape=(n_constraints, n_features), optional (default=None)
        Constraint matrix for linear inequality constraints of the form
        ``A_ineq w <= b_ineq``.

    b_ineq : array-like, shape=(n_constraints,), optional (default=None)
        Constraint vector for linear inequality constraints of the form
        ``A_ineq w <= b_ineq``.

    cv : int, cross-validation generator or Iterable, optional (default=None)
        Determines the cross-validation splitting strategy. One of:

        - ``None``, to use the default 5-fold cross-validation,
        - ``int``, to specify the number of folds.
        - ``Iterable`` yielding (train, test) splits as arrays of indices.

        For integer/``None`` inputs, :class:`KFold` is used

    n_jobs : int, optional (default=None)
        The maximum number of concurrently running jobs. The number of jobs that
        are needed is ``len(l1_ratio)`` x ``n_folds``. ``-1`` is the same as the
        number of CPU on your machine. ``None`` means ``1`` unless in a
        :obj:`joblib.parallel_backend` context.


    Attributes
    ----------
    alpha_: float
        The amount of regularization chosen by cross validation.

    alphas_: array, shape (n_l1_ratios, n_alphas)
        Alphas used by the model.

    l1_ratio_: float
        The compromise between L1 and L2 regularization chosen by cross
        validation.

    coef_ : array, shape (n_features,)
        Estimated coefficients for the linear predictor in the GLM at the
        optimal (``l1_ratio_``, ``alpha_``).

    intercept_ : float
        Intercept (a.k.a. bias) added to linear predictor.

    n_iter_ : int
        The number of iterations run by the CD solver to reach the specified
        tolerance for the optimal alpha.

    coef_path_ : array, shape (n_folds, n_l1_ratios, n_alphas, n_features)
        Estimated coefficients for the linear predictor in the GLM at every
        point along the regularization path.

    deviance_path_: array, shape(n_folds, n_alphas)
        Deviance for the test set on each fold, varying alpha.
    """

    def __init__(
        self,
        l1_ratio=0,
        P1="identity",
        P2="identity",
        fit_intercept=True,
        family: Union[str, ExponentialDispersionModel] = "normal",
        link: Union[str, Link] = "auto",
        solver="auto",
        max_iter=100,
        gradient_tol: Optional[float] = None,
        step_size_tol: Optional[float] = None,
        hessian_approx: float = 0.0,
        warm_start: bool = False,
        n_alphas: int = 100,
        alphas: Optional[np.ndarray] = None,
        min_alpha_ratio: Optional[float] = None,
        min_alpha: Optional[float] = None,
        start_params: Optional[np.ndarray] = None,
        selection: str = "cyclic",
        random_state=None,
        copy_X: bool = True,
        check_input: bool = True,
        verbose=0,
        scale_predictors: bool = False,
        lower_bounds: Optional[np.ndarray] = None,
        upper_bounds: Optional[np.ndarray] = None,
        A_ineq: Optional[np.ndarray] = None,
        b_ineq: Optional[np.ndarray] = None,
        force_all_finite: bool = True,
        cv=None,
        n_jobs: Optional[int] = None,
    ):
        self.alphas = alphas
        self.cv = cv
        self.n_jobs = n_jobs
        super().__init__(
            l1_ratio=l1_ratio,
            P1=P1,
            P2=P2,
            fit_intercept=fit_intercept,
            family=family,
            link=link,
            solver=solver,
            max_iter=max_iter,
            gradient_tol=gradient_tol,
            step_size_tol=step_size_tol,
            hessian_approx=hessian_approx,
            warm_start=warm_start,
            n_alphas=n_alphas,
            min_alpha_ratio=min_alpha_ratio,
            min_alpha=min_alpha,
            start_params=start_params,
            selection=selection,
            random_state=random_state,
            copy_X=copy_X,
            check_input=check_input,
            verbose=verbose,
            scale_predictors=scale_predictors,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            A_ineq=A_ineq,
            b_ineq=b_ineq,
            force_all_finite=force_all_finite,
        )

    def _validate_hyperparameters(self) -> None:
        if self.alphas is not None and np.any(np.asarray(self.alphas) < 0):
            raise ValueError
        l1_ratio = np.asarray(self.l1_ratio)
        if (
            not np.issubdtype(l1_ratio.dtype, np.number)
            or np.any(l1_ratio < 0)
            or np.any(l1_ratio > 1)
        ):
            raise ValueError(
                "l1_ratio must be a number in interval [0, 1]; got l1_ratio={}".format(
                    self.l1_ratio
                )
            )
        super()._validate_hyperparameters()

    def fit(
        self,
        X: ArrayLike,
        y: ArrayLike,
        sample_weight: Optional[ArrayLike] = None,
        offset: Optional[ArrayLike] = None,
    ):
        r"""
        Choose the best model along a 'regularization path' by cross-validation.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            Training data. Note that a ``float32`` matrix is acceptable and will
            result in the entire algorithm being run in 32-bit precision.
            However, for problems that are poorly conditioned, this might result
            in poor convergence or flawed parameter estimates. If a Pandas data
            frame is provided, it may contain categorical columns. In that case,
            a separate coefficient will be estimated for each category. No
            category is omitted. This means that some regularization is required
            to fit models with an intercept or models with several categorical
            columns.

        y : array-like, shape (n_samples,)
            Target values.

        sample_weight : array-like, shape (n_samples,), optional (default=None)
            Individual weights w_i for each sample. Note that, for an
            Exponential Dispersion Model (EDM), one has
            :math:`\mathrm{var}(y_i) = \phi \times v(mu) / w_i`. If
            :math:`y_i \sim EDM(\mu, \phi / w_i)`, then
            :math:`\sum w_i y_i / \sum w_i \sim EDM(\mu, \phi / \sum w_i)`,
            i.e. the mean of :math:`y` is a weighted average with weights equal
            to ``sample_weight``.

        offset: array-like, shape (n_samples,), optional (default=None)
            Added to linear predictor. An offset of 3 will increase expected
            ``y`` by 3 if the link is linear and will multiply expected ``y`` by
            3 if the link is logarithmic.
        """
        self._validate_hyperparameters()

        (
            X,
            y,
            sample_weight,
            offset,
            weights_sum,
            P1,
            P2,
        ) = self._set_up_and_check_fit_args(
            X,
            y,
            sample_weight,
            offset,
            solver=self.solver,
            force_all_finite=self.force_all_finite,
        )

        #########
        # Checks
        self._set_up_for_fit(y)
        if (
            hasattr(self._family_instance, "_power")
            and self._family_instance._power == 1.5  # type: ignore
        ):
            assert isinstance(self._link_instance, LogLink)

        l1_ratio = np.atleast_1d(self.l1_ratio)

        if self.alphas is None:
            alphas = [self._get_alpha_path(l1, X, y, sample_weight) for l1 in l1_ratio]
        else:
            alphas = np.tile(np.sort(self.alphas)[::-1], (len(l1_ratio), 1))

        if len(l1_ratio) == 1:
            self.alphas_ = alphas[0]
        else:
            self.alphas_ = np.asarray(alphas)

        lower_bounds = check_bounds(self.lower_bounds, X.shape[1], X.dtype)
        upper_bounds = check_bounds(self.upper_bounds, X.shape[1], X.dtype)

        A_ineq = copy.copy(self.A_ineq)
        b_ineq = copy.copy(self.b_ineq)

        cv = check_cv(self.cv)

        if self._solver == "cd":
            _stype = ["csc"]
        else:
            _stype = ["csc", "csr"]

        def fit_path(
            self,
            train_idx,
            test_idx,
            X,
            y,
            l1,
            alphas,
            sample_weight,
            offset,
            lower_bounds,
            upper_bounds,
            A_ineq,
            b_ineq,
        ):

            x_train, y_train, w_train = (
                X[train_idx, :],
                y[train_idx],
                sample_weight[train_idx],
            )
            w_train /= w_train.sum()

            x_test, y_test, w_test = (
                X[test_idx, :],
                y[test_idx],
                sample_weight[test_idx],
            )

            if offset is not None:
                offset_train = offset[train_idx]
                offset_test = offset[test_idx]
            else:
                offset_train, offset_test = None, None

            def _get_deviance(coef):
                mu = self._link_instance.inverse(
                    _safe_lin_pred(x_test, coef, offset_test)
                )
                return self._family_instance.deviance(y_test, mu, sample_weight=w_test)

            if (
                hasattr(self._family_instance, "_power")
                and self._family_instance._power == 1.5
            ):
                assert isinstance(self._link_instance, LogLink)

            _dtype = [np.float64, np.float32]
            start_params = initialize_start_params(
                self.start_params,
                n_cols=X.shape[1],
                fit_intercept=self.fit_intercept,
                _dtype=_dtype,
            )

            P1_no_alpha = setup_p1(P1, X, X.dtype, 1, l1)
            P2_no_alpha = setup_p2(P2, X, _stype, X.dtype, 1, l1)

            (
                x_train,
                col_means,
                col_stds,
                lower_bounds,
                upper_bounds,
                A_ineq,
                P1_no_alpha,
                P2_no_alpha,
            ) = _standardize(
                x_train,
                w_train,
                self._center_predictors,
                self.scale_predictors,
                lower_bounds,
                upper_bounds,
                A_ineq,
                P1_no_alpha,
                P2_no_alpha,
            )

            coef = self._get_start_coef(
                start_params,
                x_train,
                y_train,
                w_train,
                offset_train,
                col_means,
                col_stds,
            )

            if self.check_input:
                # check if P2 is positive semidefinite
                if not isinstance(self.P2, str):  # self.P2 != 'identity'
                    if not is_pos_semidef(P2_no_alpha):
                        if P2_no_alpha.ndim == 1 or P2_no_alpha.shape[0] == 1:
                            error = "1d array P2 must not have negative values."
                        else:
                            error = "P2 must be positive semi-definite."
                        raise ValueError(error)

            coef = self._solve_regularization_path(
                X=x_train,
                y=y_train,
                sample_weight=w_train,
                alphas=alphas,
                P2_no_alpha=P2_no_alpha,
                P1_no_alpha=P1_no_alpha,
                coef=coef,
                offset=offset_train,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
                A_ineq=A_ineq,
                b_ineq=b_ineq,
            )

            if self.fit_intercept:
                intercept_path_, coef_path_ = _unstandardize(
                    col_means, col_stds, coef[:, 0], coef[:, 1:]
                )
                assert isinstance(intercept_path_, np.ndarray)  # make mypy happy
                deviance_path_ = [
                    _get_deviance(_coef)
                    for _coef in np.concatenate(
                        [intercept_path_[:, np.newaxis], coef_path_], axis=1
                    )
                ]
            else:
                # set intercept to zero as the other linear models do
                intercept_path_, coef_path_ = _unstandardize(
                    col_means, col_stds, np.zeros(coef.shape[0]), coef
                )
                deviance_path_ = [_get_deviance(_coef) for _coef in coef_path_]

            return intercept_path_, coef_path_, deviance_path_

        jobs = (
            delayed(fit_path)(
                self,
                train_idx=train_idx,
                test_idx=test_idx,
                X=X,
                y=y,
                l1=this_l1_ratio,
                alphas=this_alphas,
                sample_weight=sample_weight,
                offset=offset,
                lower_bounds=lower_bounds,
                upper_bounds=upper_bounds,
                A_ineq=A_ineq,
                b_ineq=b_ineq,
            )
            for train_idx, test_idx in cv.split(X, y)
            for this_l1_ratio, this_alphas in zip(l1_ratio, alphas)
        )
        paths_data = Parallel(n_jobs=self.n_jobs, prefer="processes")(jobs)

        self.intercept_path_ = np.reshape(
            [elmt[0] for elmt in paths_data],
            (cv.get_n_splits(), len(l1_ratio), len(alphas[0]), -1),
        )

        self.coef_path_ = np.reshape(
            [elmt[1] for elmt in paths_data],
            (cv.get_n_splits(), len(l1_ratio), len(alphas[0]), -1),
        )

        self.deviance_path_ = np.reshape(
            [elmt[2] for elmt in paths_data],
            (cv.get_n_splits(), len(l1_ratio), len(alphas[0])),
        )

        avg_deviance = self.deviance_path_.mean(axis=0)  # type: ignore

        best_l1, best_alpha = np.unravel_index(
            np.argmin(avg_deviance), avg_deviance.shape
        )

        if len(l1_ratio) > 1:
            self.l1_ratio_ = l1_ratio[best_l1]
            self.alpha_ = self.alphas_[best_l1, best_alpha]
        else:
            self.l1_ratio_ = l1_ratio[best_l1]
            self.alpha_ = self.alphas_[best_alpha]

        P1 = setup_p1(P1, X, X.dtype, self.alpha_, self.l1_ratio_)
        P2 = setup_p2(P2, X, _stype, X.dtype, self.alpha_, self.l1_ratio_)

        # Refit with full data and best alpha and lambda
        (
            X,
            col_means,
            col_stds,
            lower_bounds,
            upper_bounds,
            A_ineq,
            P1,
            P2,
        ) = _standardize(
            X,
            sample_weight,
            self._center_predictors,
            self.scale_predictors,
            lower_bounds,
            upper_bounds,
            A_ineq,
            P1,
            P2,
        )

        start_params = initialize_start_params(
            self.start_params,
            n_cols=X.shape[1],
            fit_intercept=self.fit_intercept,
            _dtype=X.dtype,
        )

        coef = self._get_start_coef(
            start_params, X, y, sample_weight, offset, col_means, col_stds
        )

        coef = self._solve(
            X=X,
            y=y,
            sample_weight=sample_weight,
            P2=P2,
            P1=P1,  # type: ignore
            coef=coef,
            offset=offset,
            lower_bounds=lower_bounds,
            upper_bounds=upper_bounds,
            A_ineq=A_ineq,
            b_ineq=b_ineq,
        )

        if self.fit_intercept:
            self.intercept_, self.coef_ = _unstandardize(
                col_means, col_stds, coef[0], coef[1:]
            )
        else:
            # set intercept to zero as the other linear models do
            self.intercept_, self.coef_ = _unstandardize(col_means, col_stds, 0.0, coef)

        self._tear_down_from_fit()

        return self



import pkg_resources

from ._distribution import TweedieDistribution
from ._glm import GeneralizedLinearRegressor
from ._glm_cv import GeneralizedLinearRegressorCV

try:
    __version__ = pkg_resources.get_distribution(__name__).version
except Exception:
    __version__ = "unknown"

__all__ = [
    "GeneralizedLinearRegressor",
    "TweedieDistribution",
    "GeneralizedLinearRegressorCV",
]



import warnings
from abc import ABCMeta, abstractmethod
from typing import Union

import numpy as np
from scipy import special

from ._util import _asanyarray


class Link(metaclass=ABCMeta):
    """Abstract base class for Link functions."""

    @abstractmethod
    def link(self, mu):
        """Compute the link function ``g(mu)``.

        The link function links the mean, ``mu ≡ E(Y)``, to the linear predictor
        ``X * w``, i.e. ``g(mu)`` is equal to the linear predictor.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Usually the (predicted) mean.
        """
        pass

    @abstractmethod
    def derivative(self, mu):
        """Compute the derivative of the link ``g'(mu)``.

        Parameters
        ----------
        mu : array-like, shape (n_samples,)
            Usually the (predicted) mean.
        """
        pass

    @abstractmethod
    def inverse(self, lin_pred):
        """Compute the inverse link function ``h(lin_pred)``.

        Gives the inverse relationship between linear predictor,
        ``lin_pred ≡ X * w``, and the mean, ``mu ≡ E(Y)``, i.e.
        ``h(lin_pred) = mu``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        pass

    @abstractmethod
    def inverse_derivative(self, lin_pred):
        """Compute the derivative of the inverse link function ``h'(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        pass

    @abstractmethod
    def inverse_derivative2(self, lin_pred):
        """Compute second derivative of the inverse link function ``h''(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        pass


class IdentityLink(Link):
    """The identity link function ``g(x) = x``."""

    # unnecessary type hint for consistency with other methods
    def link(self, mu):
        """Return mu (identity link).

        See superclass documentation.

        Parameters
        ----------
        mu: array-like
        """
        return _asanyarray(mu)

    def derivative(self, mu):
        """Get the derivative of the identity link, a vector of ones.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like
        """
        return 1.0 if np.isscalar(mu) else np.ones_like(mu)

    # unnecessary type hint for consistency with other methods
    def inverse(self, lin_pred):
        """Compute the inverse link function ``h(lin_pred)``.

        Gives the inverse relationship between linear predictor and the mean
        ``mu ≡ E(Y)``, i.e. ``h(linear predictor) = mu``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        return _asanyarray(lin_pred)

    def inverse_derivative(self, lin_pred):
        """Compute the derivative of the inverse link function ``h'(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        return 1.0 if np.isscalar(lin_pred) else np.ones_like(lin_pred)

    def inverse_derivative2(self, lin_pred):
        """Compute second derivative of the inverse link function ``h''(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        return 0.0 if np.isscalar(lin_pred) else np.zeros_like(lin_pred)


class LogLink(Link):
    """The log link function ``log(x)``."""

    def link(self, mu):
        """Get the log of ``mu``.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like

        Returns
        -------
        numpy.ndarray
        """
        return np.log(mu)

    def derivative(self, mu):
        """Get the derivative of the log link, one over ``mu``.

        Parameters
        ----------
        mu: array-like

        Returns
        -------
        numpy.ndarray
        """
        return 1 / mu

    def inverse(self, lin_pred):
        """Get the inverse of the log link, the exponential function.

        See superclass documentation.

        Parameters
        ----------
        lin_pred: array-like

        Returns
        -------
        numpy.ndarray
        """
        return np.exp(lin_pred)

    def inverse_derivative(self, lin_pred):
        """Compute the derivative of the inverse link function ``h'(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        return np.exp(lin_pred)

    def inverse_derivative2(self, lin_pred):
        """Compute 2nd derivative of the inverse link function ``h''(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        return np.exp(lin_pred)


class LogitLink(Link):
    """The logit link function ``logit(x)``."""

    def link(self, mu):
        """Get the logit function of ``mu``.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like

        Returns
        -------
        numpy.ndarray
        """
        return special.logit(_asanyarray(mu))

    def derivative(self, mu):
        """Get the derivative of the logit link.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like

        Returns
        -------
        array-like
        """
        mu = _asanyarray(mu)
        return 1.0 / (mu * (1 - mu))

    def inverse(self, lin_pred):
        """Get the inverse of the logit link.

        See superclass documentation.

        Note: since passing a very large value might result in an output of one,
        this function bounds the output to be between ``[50*eps, 1 - 50*eps]``,
        where ``eps`` is floating point epsilon.

        Parameters
        ----------
        lin_pred: array-like

        Returns
        -------
        array-like
        """
        inv_logit = special.expit(_asanyarray(lin_pred))
        eps50 = 50 * np.finfo(inv_logit.dtype).eps
        if np.any(inv_logit > 1 - eps50) or np.any(inv_logit < eps50):
            warnings.warn(
                "Computing sigmoid function gave results too close to 0 or 1. Clipping."
            )
            return np.clip(inv_logit, eps50, 1 - eps50)
        return inv_logit

    def inverse_derivative(self, lin_pred):
        """Compute the derivative of the inverse link function ``h'(lin_pred)``.

        Parameters
        ----------
        lin_pred : array, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        ep = special.expit(_asanyarray(lin_pred))
        return ep * (1.0 - ep)

    def inverse_derivative2(self, lin_pred):
        """Compute 2nd derivative of the inverse link function ``h''(lin_pred)``.

        Parameters
        ----------
        lin_pred : array, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        ep = special.expit(_asanyarray(lin_pred))
        return ep * (1.0 - ep) * (1.0 - 2 * ep)


def catch_p(fun):
    """
    Decorate ``fun``, ensuring that the given linear predictor is compatible with the \
    relevant Tweedie power parameter.

    Parameters
    ----------
    fun: TweedieLink method

    Returns
    -------
    Callable
    """

    def _to_return(*args, **kwargs):
        with np.errstate(invalid="raise"):
            try:
                result = fun(*args, **kwargs)
            except FloatingPointError:
                raise ValueError(
                    f"Your linear predictors are not supported for p={args[0].p}. For "
                    + "negative linear predictors, consider using a log link instead."
                )
        return result

    return _to_return


class TweedieLink(Link):
    """The Tweedie link function ``x^(1-p)`` if ``p≠1`` and ``log(x)`` if ``p=1``."""

    def __new__(cls, p: Union[float, int]):
        """
        Create a new ``TweedieLink`` object.

        Parameters
        ----------
        p: scalar
        """
        if p == 0:
            return IdentityLink()
        if p == 1:
            return LogLink()
        return super(TweedieLink, cls).__new__(cls)

    def __init__(self, p):
        self.p = p

    def link(self, mu):
        """
        Get the Tweedie canonical link.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like
        """
        return _asanyarray(mu) ** (1 - self.p)

    def derivative(self, mu):
        """
        Get the derivative of the Tweedie link.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like
        """
        return (1 - self.p) * _asanyarray(mu) ** (-self.p)

    @catch_p
    def inverse(self, lin_pred):
        """
        Get the inverse of the Tweedie link.

        See superclass documentation.

        Parameters
        ----------
        mu: array-like
        """
        return _asanyarray(lin_pred) ** (1 / (1 - self.p))

    @catch_p
    def inverse_derivative(self, lin_pred):
        """Compute the derivative of the inverse Tweedie link function ``h'(lin_pred)``.

        Parameters
        ----------
        lin_pred : array-like, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        return (1 / (1 - self.p)) * _asanyarray(lin_pred) ** (self.p / (1 - self.p))

    @catch_p
    def inverse_derivative2(self, lin_pred):
        """Compute secondnd derivative of the inverse Tweedie link function \
            ``h''(lin_pred)``.

        Parameters
        ----------
        lin_pred : array, shape (n_samples,)
            Usually the (fitted) linear predictor.
        """
        result = _asanyarray(lin_pred) ** ((2 * self.p - 1) / (1 - self.p))
        result *= self.p / (1 - self.p) ** 2
        return result
